import time

from distask import util
from distask.tiggers.base import Tigger


class IntervalTigger(Tigger):
    '''每隔相同的时间会触发'''

    def __init__(self, microseconds = None, seconds=None, minutes=None, hours=None, days=None):
        self.microseconds = 0
        if microseconds: self.microseconds = microseconds 
        if seconds: self.microseconds = seconds * 1000 + self.microseconds
        if minutes: self.microseconds = minutes * 60_000 + self.microseconds
        if hours: self.microseconds = hours * 3600_000 + self.microseconds
        if days: self.microseconds = days * 86400_000 + self.microseconds
        if not self.microseconds:
            raise AttributeError("not vaild interval")

    def get_next_time(self, pre = None, limit=None):
        now = util.micro_now()
        tiggers = []
        if not pre or pre >= now:
            tiggers.append(now + self.microseconds)
            return tiggers
        for t in range(pre, now, self.microseconds):
            tiggers.append(t + self.microseconds)
        return tiggers

    def __str__(self) -> str:
        return "interval {} microseconds".format(self.microseconds)

    def __getstate__(self):  
        """Return state values to be pickled."""
        return (self.microseconds)

    def __setstate__(self, state):
        """Restore state from the unpickled state values."""  
        self.microseconds = state



import logging
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from traceback import print_exc
from typing import List, Optional

import pymongo
from pymongo import ASCENDING, DeleteOne, MongoClient, UpdateOne
from pymongo.collection import Collection
from pymongo.errors import DuplicateKeyError

from distask import util
from distask.serializers.base import Serializer
from distask.task import DeserializationError, Job
from distask.datastores.base import DataStore


class MongoDataStore(DataStore):
    def __init__(self, client: MongoClient, *, serializer: Optional[Serializer] = None,
                database: str = 'distask', 
                schedules: str = 'schedules',
                jobs: str = 'jobs',
                store_status: bool = True):
        self._client = client
        self._database = database
        self._serializer = serializer
        self._schedules = client[self._database][schedules]
        self._jobs = client[self._database][jobs]
        self._store_status = store_status

        self._create_index()

    def _create_index(self):
        self._jobs.create_index(
        [
            ("job_id", pymongo.ASCENDING),
            ("group", pymongo.ASCENDING),
            ("subgroup", pymongo.ASCENDING),
        ], unique=True)

        self._jobs.create_index(
        [
            ("next_time", pymongo.ASCENDING),
        ])

    def _reconstitute_job(self, row):
        job = Job.__new__(Job)
        states = (row['job_id'], row['next_time'], row['group'], row['subgroup'], 
            self._serializer.deserialize(row['tigger']), *self._serializer.deserialize(row['call']), row['status_last_time'])
        job.__setstate__(states)
        return job


    def get_all_jobs(self, scheduler) -> List:
        jobs: List[Job] = []
        filters = {}
        if len(scheduler._groups): filters["group"] = {"$in": scheduler._groups}
        if len(scheduler._subgroups): filters["subgroup"] = {"$in": scheduler._subgroups}
        
        rows = self._jobs.find(filters).sort([('next_time', pymongo.ASCENDING)])
        failed_job_ids = []
        for row in rows:
            try:
                job = self._reconstitute_job(row)
                jobs.append(job)
            except Exception:
                failed_job_ids.append(row['_id'])
                continue
        # Remove all the jobs we failed to restore
        if failed_job_ids:
            logging.warning('Failed to deserialize job %s, so remove it now', failed_job_ids)
            self._jobs.remove({'_id': {'$in': failed_job_ids}})
        return jobs

    def get_jobs(self, scheduler, now, limit=None) -> List:
        jobs: List[Job] = []
        filters = {"next_time": {"$lt": now}, "status_last_time": {"$lt": now}}
        if len(scheduler._groups): filters["group"] = {"$in": scheduler._groups}
        if len(scheduler._subgroups): filters["subgroup"] = {"$in": scheduler._subgroups}
        
        rows = self._jobs.find(filters).sort([('next_time', pymongo.ASCENDING)])
        if limit:
            rows = rows.limit(limit)
        failed_job_ids = []
        for row in rows:
            try:
                job = self._reconstitute_job(row)
                jobs.append(job)
            except Exception:
                failed_job_ids.append(row['_id'])
                continue
        # Remove all the jobs we failed to restore
        if failed_job_ids:
            logging.warning('Failed to deserialize job %s, so remove it now', failed_job_ids)
            self._jobs.remove({'_id': {'$in': failed_job_ids}})
        return jobs

    def add_job(self, job: Job) -> None:
        filter = {
            'job_id': job.job_id,
            'group': job.group,
            'subgroup': job.subgroup,
        }
        old = self._jobs.find_one(filter)
        if not old:
            document = {
                'job_id': job.job_id,
                'group': job.group,
                'subgroup': job.subgroup,
                'status_last_time': job.status_last_time,
                'next_time': job.next_time,
                'tigger': self._serializer.serialize(job.tigger),
                'call': self._serializer.serialize((job.func, job.args)),
            }
            self._jobs.insert_one(document)
        else:
            update = {
                'tigger': self._serializer.serialize(job.tigger),
                'call': self._serializer.serialize((job.func, job.args)),
            }
            if job.next_time < old['next_time']:
                update['next_time'] = job.next_time
            self._jobs.update_one({'_id': old['_id']}, {
                "$set": update
            })

    def modify_job(self, job, update):
        filter = {
            'job_id': job.job_id,
            'group': job.group,
            'subgroup': job.subgroup,
        }
        self._jobs.update_one(filter, {"$set": update})

    def remove_job(self, job):
        filter = {
            'job_id': job['job_id'],
            'group': job['group'],
            'subgroup': job['subgroup'],
        }
        row = self._jobs.find_one_and_delete(filter)
        try:
            return self._reconstitute_job(row)
        except Exception:
            return None

    def record_job_exec(self, status, job, duration=0, runtimes=1, excetion=''):
        if not self._store_status:
            return
        self._schedules.insert_one({
            'job_id': job.job_id,
            'group': job.group,
            'subgroup': job.subgroup,
            'status_time': util.micro_now(),
            'status': status,
            'duration': duration,
            'runtimes': runtimes,
            'excetion': excetion
        })
        return None


from dataclasses import dataclass, field
from traceback import print_exc
from distask.util import obj_to_ref, ref_to_obj
from json import dumps, loads
from typing import Any, Callable, Dict, Tuple

from distask.serializers.base import Serializer
from distask.task import DeserializationError

def callable_from_ref(ref: str) -> Callable:
    """
    Return the callable pointed to by ``ref``.

    :raises DeserializationError: if the reference could not be resolved or the looked up object is
        not callable

    """
    if ':' not in ref:
        raise ValueError(f'Invalid reference: {ref}')

    modulename, rest = ref.split(':', 1)
    try:
        obj = __import__(modulename, fromlist=[rest])
    except ImportError:
        raise LookupError(f'Error resolving reference {ref!r}: could not import module')

    try:
        for name in rest.split('.'):
            obj = getattr(obj, name)
    except Exception:
        raise DeserializationError(f'Error resolving reference {ref!r}: error looking up object')

    if not callable(obj):
        raise DeserializationError(f'{ref!r} points to an object of type '
                                   f'{obj.__class__.__qualname__} which is not callable')

    return obj

def marshal_object(obj) -> Tuple[str, Any]:
    return f'{obj.__class__.__module__}:{obj.__class__.__qualname__}', obj.__getstate__()


def unmarshal_object(ref: str, state):
    cls = callable_from_ref(ref)
    instance = cls.__new__(cls)
    instance.__setstate__(state)
    return instance

@dataclass
class JSONSerializer(Serializer):
    magic_key: str = '_distask_json'
    magic_func_key: str = '_distask_func_json'
    dump_options: Dict[str, Any] = field(default_factory=dict)
    load_options: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        self.dump_options['default'] = self._default_hook
        self.load_options['object_hook'] = self._object_hook

    @classmethod
    def _default_hook(cls, obj):
        if hasattr(obj, '__getstate__'):
            cls_ref, state = marshal_object(obj)
            return {cls.magic_key: [cls_ref, state]}
        if hasattr(obj, '__qualname__'):
            return {cls.magic_func_key: obj_to_ref(obj)}
        raise TypeError(f'Object of type {obj.__class__.__name__!r} is not JSON serializable')

    @classmethod
    def _object_hook(cls, obj_state: Dict[str, Any]):
        if cls.magic_key in obj_state:
            ref, *rest = obj_state[cls.magic_key]
            return unmarshal_object(ref, *rest)
        if cls.magic_func_key in obj_state:
            ref  = obj_state[cls.magic_func_key]
            return ref_to_obj(ref)
        return obj_state

    def serialize(self, obj) -> bytes:
        return dumps(obj, ensure_ascii=False, **self.dump_options).encode('utf-8')

    def deserialize(self, serialized: bytes):
        return loads(serialized, **self.load_options)

    def serialize_to_unicode(self, obj) -> str:
        return dumps(obj, ensure_ascii=False, **self.dump_options)

    def deserialize_from_unicode(self, serialized: str):
        return loads(serialized, **self.load_options)




from abc import ABCMeta, abstractmethod
from base64 import b64decode, b64encode
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, FrozenSet, Iterable, Iterator, List, Optional, Set, Type
from datetime import datetime, timedelta

class Serializer(metaclass=ABCMeta):
    __slots__ = ()

    @abstractmethod
    def serialize(self, obj) -> bytes:
        pass

    def serialize_to_unicode(self, obj) -> str:
        return b64encode(self.serialize(obj)).decode('ascii')

    @abstractmethod
    def deserialize(self, serialized: bytes):
        pass

    def deserialize_from_unicode(self, serialized: str):
        return self.deserialize(b64decode(serialized))



import random
import time

from distask.locks.base import BaseLock
from distask import util
from redlock import RedLock
from redlock.lock import ReentrantRedLock


DEFAULT_RLLOCK_NAME = "default_distributed_lock"

class RLLock(BaseLock):
    _lock = None
    def __init__(self, **kwargs):
        if kwargs.pop("reentrant", None):
            self._lock = ReentrantRedLock(DEFAULT_RLLOCK_NAME, **kwargs)
        else:
            self._lock = RedLock(DEFAULT_RLLOCK_NAME, **kwargs)

    def lock(self, timeout: float = -1):
        if timeout == -1:
            return self._lock.acquire()
        now = util.micro_now()
        while util.micro_now() - now < timeout:
            if self._lock.acquire():
                return True
            time.sleep(random.uniform(0, util.RETRY_DELAY/1000))
        return False

    def unlock(self):
        return self._lock.release()



import logging
import time, os, sys
from datetime import date, datetime, timezone
from os import PRIO_PGRP

from pymongo.mongo_client import MongoClient

base_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
print("base_path ==", base_path)
sys.path.append(base_path)
sys.path.append(base_path + '/..')
sys.path.append(os.pardir)  # 为了导入父目录的文件而进行的设定

from distask import create_scheduler, task, register_job
from distask import util
from distask.datastores.mongodb import MongoDataStore
from distask.events import EVENT_SCHEDULER_START
from distask.locks.rl_lock import RLLock
from distask.schedulers.background import BackgroundScheduler
from distask.schedulers.base import Scheduler
from distask.serializers.json import JSONSerializer
from distask.serializers.pickle import PickleSerializer
from distask.tiggers.interval import IntervalTigger


def test00(times, aa=None, bb=None, *args):
    print("test0 ======================")
    time.sleep(0.1)
    # return True

def test_exception(times, aa=None, bb=None, *args):
    print("test0 ======================")
    a = 1 / 0
    # return True

def testcron(times, aa=None, bb=None):
    print("cron====================== ", util.time_now())

    # exit(0)
logging.basicConfig(level=logging.DEBUG)
tigger = IntervalTigger(seconds=100)

# job = task.Job()

# serialize = PickleSerializer()
serialize = JSONSerializer()
client = MongoClient("mongodb://admin:123456@192.168.99.27:27017")
mongodb = MongoDataStore(client, serializer=serialize)
# connection_details=[
#     {'host': '192.168.99.27', 'port': 6379, 'db': 0},
# ]
connection_details=[
    {'host': '127.0.0.1', 'port': 6379, 'db': 0},
]
lock = RLLock(reentrant=True, connection_details=connection_details, ttl=10_000)
# scheduler = Scheduler(store=mongodb, lock=lock, group="", limit=1, maxwait=5)
# scheduler = Scheduler(store=mongodb, lock=lock, groups=['test'], limit=1, maxwait=5)
client_data = {
    "t": "mongo",
    "args": ["mongodb://admin:123456@192.168.99.27:27017"]
}
lock_data = {
    "t": "rllock",
    "reentrant":True, 
    "connection_details":connection_details, 
    "ttl":10_000
}
scheduler = create_scheduler(client_data, lock_data, serialize="pickle", groups=['test'], limit=1, maxwait=5)

@register_job(scheduler, "interval", args=(12, 123), group="test", subgroup="ssss", seconds=3)
def test1(times, aa=None, bb=None, *args):
    print("test---------------------", util.time_now())

job = task.Job(test00, "interval", (12, 123), group="test", subgroup="ssss", seconds=3)
scheduler.add_job(job)

job = task.Job(test_exception, "interval", (12, 123), group="test", subgroup="ssss", seconds=3)
scheduler.add_job(job)

def job_execute(event):

    if event.code == EVENT_SCHEDULER_START:
        print("start success")
scheduler.add_listener(job_execute, EVENT_SCHEDULER_START)
scheduler.start()
while True:
    print("sleep!!!!!!!")
    time.sleep(5)
scheduler.shutdown()
# jobs = scheduler._get_jobs()
# print(jobs)
# for job in jobs:
#     if job.func:
#         tiggers = job.tigger.get_next_time(job.next_time * 1000)
#         print("tiggers len ==", len(tiggers))
#         print(job.tigger)
#         job.func(tiggers, *job.args)




import logging
from abc import ABC, abstractmethod
from datetime import date, datetime, timezone
from threading import Event, RLock

from distask.events import (EVENT_ALL, EVENT_JOB_ADDED, EVENT_JOB_REMOVED,
                            EVENT_SCHEDULER_PAUSED, EVENT_SCHEDULER_RESUMED,
                            EVENT_SCHEDULER_SHUTDOWN, EVENT_SCHEDULER_START,
                            JobEvent, SchedulerEvent)
from distask.task import Job
from distask import util

DEFAULT_MAX_JOB = 10
DEFAULT_MAX_WAIT = 60

STATE_STOPPED = 0
STATE_RUNNING = 1
STATE_PAUSED = 2

class Scheduler(ABC):
    ''''''
    def __init__(self, store=None, lock=None, limit=DEFAULT_MAX_JOB, groups=[], subgroups=[], maxwait=DEFAULT_MAX_WAIT) -> None:
        self._store = store
        self._limit = limit
        self._lock = lock
        assert type(groups) == list, "groups must be list"
        assert type(subgroups) == list, "subgroups must be list"
        self._groups = groups
        self._subgroups = subgroups
        self._maxwait = maxwait
        self._jobs = []
        self._event = None

        self.state = STATE_STOPPED
        self._listeners_lock = RLock()
        self._listeners = []

    def _process_jobs(self):
        if self.state == STATE_PAUSED:
            logging.debug('Scheduler is paused -- not processing jobs')
            return None

        logging.debug('Looking for jobs to run')
        now = util.micro_now()
        next_wakeup_time = now + 6_000
        jobs = []

        with self._lock.create_lock() as succ:
            if not succ:
                return 0.1
            try:
                jobs = self._get_jobs(self._limit)
            except Exception as e:
                import traceback
                traceback.print_exc()
                logging.warning('get job occur error')
                pass
            
            # update task status
            for job in jobs:
                changes = {
                    "status_last_time": now + job.deal_max_time,
                }
                self._store.modify_job(job, changes)
        for job in jobs:
            try:
                start = util.micro_now()
                tiggers = job.get_next_time()
                job.next_time, job.close_now = job.call_func(tiggers)
                self._store.record_job_exec("success", job, util.micro_now() - start, len(tiggers))
            except Exception as e:
                job.next_time = tiggers[-1]
                self._store.record_job_exec("failed", job, util.micro_now() - start, 0, str(e))
                logging.warning("run job %s catch exception(%s)" % (job.job_id, e))

        with self._lock.create_lock(60_000) as succ:
            for job in jobs:
                if job.is_only_once() or job.next_time == 0 or job.close_now:
                    self._store.remove_job(job)
                else:
                    changes = {
                        "status_last_time": 0,
                        "next_time": job.next_time
                    }
                    next_wakeup_time = min(job.next_time, next_wakeup_time)
                    self._store.modify_job(job, changes)

        wait_microseconds = next_wakeup_time - now
        if (len(jobs) == self._limit and wait_microseconds > 60_000) or wait_microseconds < 0:
            wait_microseconds = 0
        return min(self._maxwait, wait_microseconds / 1000)

    def _get_jobs(self, limit=None):
        now = util.micro_now()
        jobs = self._store.get_jobs(self, now, limit)
        return jobs

    def get_all_jobs(self, limit=None):
        jobs = self._store.get_all_jobs(self)
        return jobs

    def add_job(self, job: Job):
        try:
            self._store.add_job(job)
            self._dispatch_event(JobEvent(EVENT_JOB_ADDED, job.job_id))
        except Exception as e:
            logging.warning("add job failed reason by '%s'" % str(e))
            pass

    def remove_job(self, job):
        remove_job = self._store.remove_job({
            'job_id': job.job_id,
            'group': job.group,
            'subgroup': job.subgroup
        })

        event = JobEvent(EVENT_JOB_REMOVED, job['job_id'])
        self._dispatch_event(event)

        logging.info('Removed job %s', job['job_id'])

    def start(self, paused=False, ready=True):
        if self._event is None or self._event.is_set():
            self._event = Event()

        self.state = STATE_PAUSED if paused else STATE_RUNNING
        self._dispatch_event(SchedulerEvent(EVENT_SCHEDULER_START))
        if ready:
            self._main_loop()

    def shutdown(self, wait=True):
        self._event.set()
        self._dispatch_event(SchedulerEvent(EVENT_SCHEDULER_SHUTDOWN))


    def pause(self):
        if self.state == STATE_STOPPED:
            return
        elif self.state == STATE_RUNNING:
            self.state = STATE_PAUSED
            logging.info('Paused scheduler job processing')
            self._dispatch_event(SchedulerEvent(EVENT_SCHEDULER_PAUSED))

    def resume(self):
        if self.state == STATE_STOPPED:
            return
        elif self.state == STATE_PAUSED:
            self.state = STATE_RUNNING
            logging.info('Resumed scheduler job processing')
            self._dispatch_event(SchedulerEvent(EVENT_SCHEDULER_RESUMED))
            self.wakeup()

    def _main_loop(self):
        wait_seconds = 0
        while self.state != STATE_STOPPED:
            self._event.wait(wait_seconds)
            self._event.clear()
            try:
                wait_seconds = self._process_jobs()
            except KeyboardInterrupt:
                raise
            except Exception as e:
                logging.warning("process job exception '%s'" % str(e))

    def wakeup(self):
        self._event.set()


    def add_listener(self, callback, mask=EVENT_ALL):
        with self._listeners_lock:
            self._listeners.append((callback, mask))

    def remove_listener(self, callback):
        with self._listeners_lock:
            for i, (cb, _) in enumerate(self._listeners):
                if callback == cb:
                    del self._listeners[i]

    def _dispatch_event(self, event):
        with self._listeners_lock:
            listeners = tuple(self._listeners)

        for cb, mask in listeners:
            if event.code & mask:
                try:
                    cb(event)
                except BaseException:
                    logging.exception('Error notifying listener')



from abc import ABC,abstractmethod
from typing import List

class DataStore(ABC):
    ''''''

    @abstractmethod
    def get_jobs(self, scheduler, now, limit=None) -> List:
        return None

    @abstractmethod
    def get_all_jobs(self, scheduler) -> List:
        return None

    @abstractmethod
    def modify_job(self, job, update):
        return None

    @abstractmethod
    def remove_job(self, job):
        return None

    @abstractmethod
    def record_job_exec(self, status, job, duration=0, runtimes=1, excetion=None):
        return None


from setuptools import setup, find_packages
from os import path
here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="distask",
    version="0.1.2",
    keywords="distribute scheduling cron",
    description="a distribute task scheduler ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT Licence",
    author="tickbh",
    author_email="tickdream125@hotmail.com",
    packages=[
        'distask', 
        'distask/locks', 
        'distask/datastores', 
        'distask/schedulers', 
        'distask/serializers', 
        'distask/tiggers', 
        'distask/tiggers/cron'
    ],
    # py_modules=['distask/*', 'distask/locks/*'],
    platforms="any",
    python_requires='>=3.6',
    include_package_data=True,
    install_requires=[
        'redlock',
        'pymongo'
    ],
    extras_require={
        'extra': ['dill']
    }

)


from distask.serializers.json import JSONSerializer
from distask.serializers.pickle import PickleSerializer
from distask.task import Job
from distask.locks.base import BaseLock
from distask.datastores.base import DataStore
from distask.schedulers.base import Scheduler
from distask.schedulers.background import BackgroundScheduler

def register_job(scheduler: Scheduler, *args, **kwargs) -> callable:
    def wrapper_register_job(func):
        job = Job(func, *args, **kwargs)
        scheduler.add_job(job)
        return func
    return wrapper_register_job

def create_scheduler(client_data, lock_data, serialize="json", **kwargs):
    serialize =  PickleSerializer() if serialize == "pickle" else JSONSerializer()
    client = None
    store = None
    if isinstance(client_data, DataStore):
        store = client_data
    else:
        cliet_type = client_data.pop("t")
        if cliet_type == "mongo":
            try:
                from pymongo import MongoClient
                from distask.datastores.mongodb import MongoDataStore
                args = client_data.pop("args", [])
                client = MongoClient(*args, **client_data)
                store = MongoDataStore(client, serializer=serialize)
            except:
                raise
    assert store, "must store client exist"
    lock = None
    if isinstance(lock_data, BaseLock):
        lock = lock_data
    else:
        lock_type = lock_data.pop("t")
        if lock_type == "rllock":
            try:
                from distask.locks.rl_lock import RLLock
                args = lock_data.pop("args", [])
                lock = RLLock(*args, **lock_data)
            except:
                raise

    assert lock, "must lock exist"

    scheduler = Scheduler(store=store, lock=lock, **kwargs)
    return scheduler